/*  Connected components in the grid.  -- for 128x128 array	*/
/*  Uses start-bit scans.  Setup and solution only --		*/
/*  no output or control routines.				*/
/*  File:  grccsw.cs						*/

/*  Axis convention:						*/
/*	   0     1       --- y --->              YMX-1		*/
/*	+-----+-----+-----+-----+-----+- . . . -+-----+		*/
/*    0	|     |     |     |     |     |         |     |		*/
/*	+-----+-----+-----+-----+-----+- . . . -+-----+		*/
/*    1	|     |     |     |     |     |         |     |		*/
/*  |	+-----+-----+-----+-----+-----+- . . . -+-----+		*/
/*  x	|     |     |     |     |     |         |     |		*/
/*  |	+-----+-----+-----+-----+-----+- . . . -+-----+		*/
/*  v	:     :     :     :     :     :         :     :		*/
/*	+-----+-----+-----+-----+-----+- . . . -+-----+		*/
/* XMX-1|     |     |     |     |     |         |     |		*/
/*	+-----+-----+-----+-----+-----+- . . . -+-----+		*/

#include <stdio.h>
#include <stdlib.h>
#include <cscomm.h>

#define TRUE  1
#define FALSE 0

/* For scaling random integers to [0..100] */
#define Scale 21474836

/* Grid sizes */

#define XMX  128
#define YMX  128

/* Shape for parallel data */

shape [XMX][YMX]cells;	/* [128][128]	*/

/* Parallel vars for cell data */

int:cells cc,	/* Component number	*/
	  cw,	/* Working component	*/
	   x,	/* x coord (axis 0)	*/
	   y;	/* y coord (axis 1)	*/

/* Parallel vars for barrier data */

bool:cells nb,	/* Northern barrier	*/
	   eb,	/* Eastern barrier	*/
	   wb,	/* Western barrier	*/
	   sb;  /* Southern barrier	*/


/* Setup random test with probability of barrier = p/100. */
/* Returns total number of barriers installed [nb + wb]	  */
/* Expects 0 <= p <= 100.				  */
/* Must be called with `cells' as current shape.	  */

int setup(int p)
 { int nc, wc;	/* Barrier counts (north, west) */
   x = pcoord(0);  y = pcoord(1);	/* Initialize	*/
   nb = (x == 0);	/* Northern edge */
   wb = (y == 0);	/* Western edge  */
   /* Add random barriers on North and West */
   where ( prand()/Scale <= p ) nb = TRUE;
   where ( prand()/Scale <= p ) wb = TRUE;
   /* Make Eastern and Southern barriers consistent (uses torus) */
   sb = from_torus_dim(&nb, 0, 1);
   eb = from_torus_dim(&wb, 1, 1);
   where (nb) nc = (+= (int:current) 1);	/* Count N barriers */
   where (wb) wc = (+= (int:current) 1);	/* Count W barriers */
   return (nc + wc);
 }


/* Calculate connected components.		    */
/* Returns no. of cycles, not counting final test.  */
/* Must be called with `cells' as current shape.    */

int gccs()
 { int nc;	/* Counts cycles */
   cc = y + YMX*x;		/* Row-major position (unique)	*/
   for ( nc = 0;  gccx();  nc++) ;	/* Repeat until solved  */
   return nc;
 }

/*
int clr_number(int:cells clr)
 { int nc /*, nu, ok, bi, bj, bn */;  /* ES - 13 May 1995 */
   nc = (>?= clr);              /* No. of colors used     
   return nc;
 }
*/

/* Perform one cycle of sweeps, test for changes.		*/
/* Updates component numbers in cc.			       */
/* Returns test result, TRUE indicates one or more changes.	*/
/* Must be called with `cells' as current shape. 		*/

int gccx()
 { int r;	/* For result */
   cw = cc;
   cw = scan(cw, 1, CMC_combiner_min, CMC_upward,
	            CMC_start_bit, &wb, CMC_inclusive);
   cw = scan(cw, 1, CMC_combiner_min, CMC_downward,
	            CMC_start_bit, &eb, CMC_inclusive);
   cw = scan(cw, 0, CMC_combiner_min, CMC_upward,
	            CMC_start_bit, &nb, CMC_inclusive);
   cw = scan(cw, 0, CMC_combiner_min, CMC_downward,
	            CMC_start_bit, &sb, CMC_inclusive);
   r = (|= (cw != cc));		/* Any changes? */
   cc = cw;			/* Update component numbers */ 
   return r;
 }


/*  Modified 13 May 1995 by ES - see comments on gccs_growing  below */
int gccx_growing(int:current ccfin, int cla, int *cy, int:current root_clr)
 { int r;	/* For result */
   int:current temp;
   printf ("\n***** I am in GCCX now *****\n");
   cw = cc;
   cw = scan(cw, 1, CMC_combiner_min, CMC_upward,
	            CMC_start_bit, &wb, CMC_inclusive);
   cw = scan(cw, 1, CMC_combiner_min, CMC_downward,
	            CMC_start_bit, &eb, CMC_inclusive);
   cw = scan(cw, 0, CMC_combiner_min, CMC_upward,
	            CMC_start_bit, &nb, CMC_inclusive);
   cw = scan(cw, 0, CMC_combiner_min, CMC_downward,
	            CMC_start_bit, &sb, CMC_inclusive);
   r = (|= (cw != cc));		/* Any changes? */
   cc = cw;			/* Update component numbers */
   printf ("\n**** I am trying grcolor now ***\n");
   temp = grcolor_growing (cc, ccfin, cla, cy, root_clr);
   return r;
 }


/*  Modified by ES - 13 May 1995 */
/*  ccfin is final labelling, root_clr is corresponding coloring */
/*  cla is number of colors allowed for animation */
/*  Note that we don't use `cc' for argument, cc is an external pvar */

int gccs_growing(int:current ccfin, int cla, int *cy, int:current root_clr)
 { int nc;	/* Counts cycles */
   printf ("\n**** I am in GCCS NOW! ****\n");
   cc = y + YMX*x;		/* Row-major position (unique)	External! */
   printf ("\n**** Trying GCCX ****\n");
   for ( nc = 0;  gccx_growing(ccfin, cla, cy, root_clr);  nc++) ;
						/* Repeat until solved  */
   return nc;
 }


